import re
from enum import Enum
from typing import List, NamedTuple, Tuple

"""
Filter is a character sequence of format:
    <name>:<values>

where values can be separated by comma.
Examples:
    1. tags:important  => one value
    2. tags:important,paid => two values
    3. tags:invoice, paid => two values

Notice that in example 3. there is a white space after comma.

Filters may contain one single value composed from multiple
words, but in such case you need to include them single
or double quotes.
Examples:
    1. breadcrumb:"My Documents"
    2. breadcrumb:'My bills'
"""


def first_filter_beg_pos(text: str) -> int | None:
    """Returns the beginning position of the first filter"""
    first_column_pos = text.find(':')

    if first_column_pos < 0:
        return None

    beg_pos = first_column_pos - 1

    while beg_pos >= 0:
        if re.match(r'\w', text[beg_pos]):
            beg_pos -= 1
        else:
            break

    return beg_pos + 1


def first_filter_end_pos(text: str) -> int | None:
    """Returns the end position of the first filter"""
    if text is None:
        return None

    first_column_pos = text.find(':')

    if first_column_pos < 0:
        return None

    text_len = len(text)
    end_pos = first_column_pos + 1
    pattern = r'[\w\, ]'
    at_least_one_comma = False
    with_quotes = False

    if text[end_pos] in "\'\"":
        end_pos += 1  # skip initial quotes
        with_quotes = True

    while end_pos < text_len:
        if re.match(pattern, text[end_pos]):
            if text[end_pos] == ' ':
                if at_least_one_comma or with_quotes:
                    end_pos += 1
                    continue
                else:
                    break
            if text[end_pos] == ',':
                at_least_one_comma = True
            end_pos += 1
        else:
            if text[end_pos] in "\'\"":
                end_pos += 1
            break

    return end_pos - 1


def first_filter_pos(text: str) -> Tuple[int, int] | None:
    first_column_pos = text.find(':')

    if first_column_pos < 0:
        return None

    beg_pos = first_filter_beg_pos(text)
    end_pos = first_filter_end_pos(text)

    return beg_pos, end_pos


def extract_free_text(text: str) -> str | None:
    pos = first_filter_pos(text)
    if pos is None:
        return text

    result = text[:pos[0]] + text[pos[1] + 1:]
    stripped_result = result.strip()

    if len(stripped_result) == 0:
        return None

    return stripped_result


def extract_filters(text: str) -> List[str]:
    # TODO: support multiple filters
    result = []
    pos = first_filter_pos(text)
    if pos is None:
        return result

    result.append(text[pos[0]:pos[1] + 1])

    return result


class FreeTextQuery(NamedTuple):
    text: str | None

    def __str__(self):
        if not self.text:
            return ''

        return self.text


class FilterQueryOP(Enum):
    # filter query operation
    AND = 1
    OR = 2


class FilterQuery:
    text: str  # original filter text
    name: str
    values: List[str]
    # which operation should the filter perform on the
    # values? Should it be AND? Should it be OR?
    op: FilterQueryOP = FilterQueryOP.AND  # AND | OR

    def __init__(self, text: str):
        self.text = text
        self.name, self.values = text.split(':')
        self.values = [v.strip() for v in self.values.split(',')]


class Query:
    original_query: str
    free_text: FreeTextQuery
    filters: List[FilterQuery]

    def __init__(self, query: str):
        self.original_query = query
        self.free_text = FreeTextQuery(extract_free_text(query))
        self.filters = [
            FilterQuery(filter_q) for filter_q in extract_filters(query)
        ]

    def get_filters_by(self, name: str) -> List[FilterQuery]:
        return [f for f in self.filters if f.name == name]

    def __repr__(self):
        return f"Query(free_text='{self.free_text}', filters={self.filters})"


class SearchQuery:
    query: Query

    def __init__(self, entity, query: str):
        self.entity = entity
        self.query = Query(query)

    def __str__(self):
        return f"SearchQuery(query={self.query}, entity={self.entity})"

    def __repr__(self):
        return f"SearchQuery(query={self.query}, entity={self.entity})"



from typing import Optional

from typing_extensions import Annotated

from .field import IdField, KeywordField, NumericField, TextField

IdPrimary = Annotated[int, IdField(primary_key=True)]

IdStrPrimary = Annotated[str, IdField(primary_key=True)]
IdStr = Annotated[Optional[str], IdField()]
OptionalStrId = Annotated[Optional[str], IdField()]

Text = Annotated[str, TextField()]
OptionalText = Annotated[Optional[str], TextField()]

OptionalNumeric = Annotated[Optional[int], NumericField()]

Keyword = Annotated[str, KeywordField()]
OptionalKeyword = Annotated[Optional[list[str]], KeywordField()]



from pydantic import BaseModel

from salinic.query import SearchQuery


class Base:
    def __init__(self, client, schema):
        self.client = client
        self.schema = schema

    def search(self, sq: SearchQuery):
        pass


class IndexRO(Base):
    pass


class IndexRW(Base):
    def add(self, model: BaseModel):
        self.client.add(model)

    def remove(self, docid: str):
        self.client.remove(docid)



import requests
from pydantic import BaseModel

from salinic.url import URL


class Base:
    def __init__(self, url: URL):
        self._url = url


class ClientRW(Base):
    def add(self, model: BaseModel):
        # change data specific for add
        data = {}
        data['add'] = {'doc': model.model_dump()}
        return requests.post(self.http_url, json=data)

    def remove(self, **kwargs):
        # change data specific for delete
        data = {}
        data['delete'] = kwargs
        return requests.post(self.http_url, json=data)

    @property
    def http_url(self) -> str:
        """Returns the http url for adding/updating/deleting documents

        As of solr9 same http url is used for all CRUD operations on
        the documents.
        Example of http url to be returned:
            http://localhost:8983/solr/papermerge_index/update/json/docs
        """
        s = "http"
        if self._url.scheme == 'solrs':
            s += "s"

        s += f"://{self._url.host}"
        if self._url.port is not None:
            s += f":{self._url.port}/solr/"
        else:
            s += "/solr/"

        s += f"{self._url.index}/update/json/docs"

        return s


ClientRO = ClientRW



import re
from enum import Enum
from typing import Optional

from pydantic import BaseModel


class SchemeEnum(str, Enum):
    xapian = 'xapian'
    solr = 'solr'
    solrs = 'solrs'
    # solr secure, means solr over secure http i.e. (https)


class URL(BaseModel):
    scheme: SchemeEnum
    """search engine backend name, such as ``xapian``, ``solr``,
    ``solrs`` (solr secure)
    """

    username: Optional[str]

    password: Optional[str]

    host: Optional[str]

    port: Optional[int]
    "integer port number"

    index: Optional[str]
    "index name"


def make_url(name_or_url: str | URL) -> URL:
    if isinstance(name_or_url, str):
        return _parse_url(name_or_url)
    elif not isinstance(name_or_url, URL):
        raise ValueError(
            f"Expected string or URL object, got {name_or_url!r}"
        )
    else:
        return name_or_url


def _parse_url(name: str) -> URL:
    pattern = re.compile(
        r"""
            (?P<scheme>[\w\+]+)://
            (?:
                (?P<username>[^:/]*)
                (?::(?P<password>[^@]*))?
            @)?
            (?:
                (?:
                    \[(?P<ipv6host>[^/\?]+)\] |
                    (?P<ipv4host>[^/:\?]+)
                )?
                (?::(?P<port>[^/\?]*))?
            )?
            (?:/(?P<index>[^\?]*))?
            """,
        re.X,
    )

    m = pattern.match(name)
    if m is not None:
        components = m.groupdict()

        ipv4host = components.pop("ipv4host")
        ipv6host = components.pop("ipv6host")
        components["host"] = ipv4host or ipv6host

        if components["port"]:
            components["port"] = int(components["port"])

        return URL(**components)

    else:
        raise ValueError(
            "Could not parse URL from string '%s'" % name
        )



import importlib

from salinic.query import SearchQuery


class IndexBase:
    def __init__(self, engine, schema):
        self.engine = engine
        self.schema = schema

    def search(self, sq: SearchQuery):
        return self.backend.search(sq)


class IndexRO(IndexBase):

    def __init__(self, engine, schema):
        super().__init__(engine, schema)
        self.backend = get_ro_index_backend(engine, schema)


class IndexRW(IndexBase):
    def __init__(self, engine, schema):
        super().__init__(engine, schema)
        self.backend = get_rw_index_backend(engine, schema)

    def add(self, entity):
        self.backend.add(entity)

    def remove(self, docid: str):
        self.backend.remove(docid)


def get_ro_client_backend_class(engine):
    module_full_path = f'salinic.backends.{engine.url.scheme}.client'
    module = importlib.import_module(module_full_path)

    return module.ClientRO


def get_rw_client_backend_class(engine):
    module_full_path = f'salinic.backends.{engine.url.scheme}.client'
    module = importlib.import_module(module_full_path)

    return module.ClientRW


def get_ro_index_backend_class(engine):
    module_full_path = f'salinic.backends.{engine.url.scheme}.index'
    module = importlib.import_module(module_full_path)

    return module.IndexRO


def get_rw_index_backend_class(engine):
    module_full_path = f'salinic.backends.{engine.url.scheme}.index'
    module = importlib.import_module(module_full_path)

    return module.IndexRW


def get_ro_index_backend(engine, schema):
    ClientKlass = get_ro_client_backend_class(engine)
    client = ClientKlass(engine.dsn)
    IndexKlass = get_ro_index_backend_class(engine)

    return IndexKlass(client, schema)


def get_rw_index_backend(engine, schema):
    ClientKlass = get_rw_client_backend_class(engine)
    client = ClientKlass(engine.url)
    IndexKlass = get_rw_index_backend_class(engine)
    return IndexKlass(client, schema)



import json

import xapian
from pydantic import BaseModel

from salinic.field import Field, IdField, KeywordField, TextField
from salinic.query import SearchQuery
from salinic.utils import first


class IndexRO:

    def __init__(self, client, schema, language="en"):
        self.client = client
        self._schema = schema
        self._language = language
        self._termgenerator = xapian.TermGenerator()
        self._termgenerator.set_stemmer(xapian.Stem(language))
        self._queryparser = xapian.QueryParser()
        self._queryparser.set_stemmer(xapian.Stem(language))
        self._queryparser.set_stemming_strategy(self._queryparser.STEM_SOME)

    def search(self, sq: SearchQuery):
        results = []

        if str(sq.query.free_text):
            query = self._queryparser.parse_query(
                str(sq.query.free_text)
            )
        else:
            query = xapian.Query('')

        for name, field in sq.entity.model_fields.items():
            self._queryparser.add_prefix(name.upper(), name.upper())
            filter_queries = sq.query.get_filters_by(name)
            for qf in filter_queries:
                xapian_fqueries = [
                    xapian.Query(qf.name.upper() + filter_value.lower())
                    for filter_value in qf.values
                ]
                # apply AND operator (OR operator) on filter queries
                xa_fq = xapian.Query(xapian.Query.OP_AND, xapian_fqueries)
                # combine free text query with filter queries
                query = xapian.Query(
                    xapian.Query.OP_FILTER,
                    query,
                    xa_fq
                )

        enquire = xapian.Enquire(self.client.db)

        enquire.set_query(query)

        for match in enquire.get_mset(0, 10):
            fields = json.loads(match.document.get_data().decode('utf8'))
            results.append(sq.entity(**fields))

        return results


class IndexRW:

    def __init__(self, client, schema, language="en"):
        self.client = client
        self._schema = schema
        self._language = language
        self._termgenerator = xapian.TermGenerator()
        self._termgenerator.set_stemmer(xapian.Stem(language))
        self._queryparser = xapian.QueryParser()
        self._queryparser.set_stemmer(xapian.Stem(language))
        self._queryparser.set_stemming_strategy(self._queryparser.STEM_SOME)

    def add(self, entity: BaseModel):
        doc = xapian.Document()
        self._termgenerator.set_document(doc)

        primary_key_name = None

        for name, field in entity.model_fields.items():
            field_instance = first(field.metadata)
            value = entity.get_field_value(name)
            if isinstance(value, Field):
                insert_value = value.default
            else:
                insert_value = value
            if not insert_value:
                continue
            prefix = name.upper()

            if isinstance(field_instance, TextField):
                index_text_field(self._termgenerator, insert_value, prefix)
            elif isinstance(field_instance, KeywordField):
                index_keyword_field(
                    self._termgenerator,
                    insert_value,
                    prefix
                )
            elif isinstance(field_instance, IdField):
                if isinstance(insert_value, str):
                    index_id_field(self._termgenerator, insert_value, prefix)

            id_field = first(field.metadata, lambda x: type(x) is IdField)

            if id_field and id_field.primary_key:
                primary_key_name = name

        doc.set_data(
            json.dumps(entity.model_dump())
        )

        if not primary_key_name:
            raise ValueError("No primary field defined")

        identifier = getattr(entity, primary_key_name)
        idterm = f"{primary_key_name.upper()}{identifier}"
        doc.add_boolean_term(idterm)

        self.client.replace_document(idterm, doc)

    def remove(self, docid: str):
        self.client.delete_document(docid)

    def search(self, sq: SearchQuery):
        results = []

        if str(sq.query.free_text):
            query = self._queryparser.parse_query(
                str(sq.query.free_text)
            )
        else:
            query = xapian.Query('')

        for name, field in sq.entity.model_fields.items():
            self._queryparser.add_prefix(name.upper(), name.upper())
            filter_queries = sq.query.get_filters_by(name)
            for qf in filter_queries:
                xapian_fqueries = [
                    xapian.Query(qf.name.upper() + filter_value.lower())
                    for filter_value in qf.values
                ]
                # apply AND operator (OR operator) on filter queries
                xa_fq = xapian.Query(xapian.Query.OP_AND, xapian_fqueries)
                # combine free text query with filter queries
                query = xapian.Query(
                    xapian.Query.OP_FILTER,
                    query,
                    xa_fq
                )

        enquire = xapian.Enquire(self.client.db)

        enquire.set_query(query)

        for match in enquire.get_mset(0, 10):
            fields = json.loads(match.document.get_data().decode('utf8'))
            results.append(sq.entity(**fields))

        return results


def index_text_field(
    term_generator: xapian.TermGenerator,
    insert_value: str,
    prefix: str,
    weight: int = 1
):
    term_generator.index_text(
        insert_value,
        weight,
        prefix  # the prefix
    )

    # index field without prefix for general search
    term_generator.index_text(insert_value)
    term_generator.increase_termpos()


def index_id_field(
    term_generator: xapian.TermGenerator,
    insert_value: str,
    prefix: str,
    weight: int = 1
):
    doc = term_generator.get_document()
    id_as_term = insert_value.replace('-', '')
    doc.add_boolean_term(prefix + id_as_term)

    term_generator.index_text(id_as_term, weight, prefix)
    term_generator.index_text(id_as_term)
    term_generator.increase_termpos()


def index_keyword_field(
    term_generator: xapian.TermGenerator,
    insert_value: str,
    prefix: str
):
    doc = term_generator.get_document()
    if isinstance(insert_value, str):
        doc.add_boolean_term(
            prefix + insert_value
        )
        doc.add_boolean_term(
            prefix + insert_value.lower()
        )

    if isinstance(insert_value, list):
        for value in insert_value:
            doc.add_boolean_term(
                prefix + value
            )
            doc.add_boolean_term(
                prefix + value.lower()
            )



from typing import NamedTuple


class Field(NamedTuple):
    primary_key: bool = False
    store: bool = True
    index: bool = True
    general_search: bool = False
    default: any = None
    multi_value: bool = False
    multi_lang: bool = False


class KeywordField(Field):
    pass


class TextField(Field):
    pass


class NumericField(Field):
    pass


class IdField(Field):
    pass


class UUIDField(Field):
    pass



from pydantic import BaseModel, ConfigDict, model_serializer

from .field import Field, IdField
from .utils import first


class Schema(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_serializer()
    def model_ser(self):
        result = {}
        for name, field in self.model_fields.items():
            default_value = getattr(self, name)
            if isinstance(default_value, Field):
                result[name] = field.default.default
            else:
                result[name] = default_value

        return result

    @property
    def pk(self):
        """returns the value of the primary key field"""
        primary_key_name = self.pk_name

        if primary_key_name is None:
            raise ValueError("No primary field defined")

        result = getattr(self, primary_key_name)
        if result:
            return result.replace('-', '')

        return result

    @property
    def pk_name(self):
        """returns the name of the primary key field"""
        for name, field in self.model_fields.items():
            id_field = first(field.metadata, lambda x: type(x) is IdField)
            if id_field and id_field.primary_key:
                return name

        return None

    def get_field_value(self, field_name):
        """Returns the field value to actually be stored on the index

        Value actually stored in the index may differ from the one in the
        model. Let's consider an example when index schema has a field
        composed of a list of colored tags (tag has a name and a color):

            from salinic import IdField, KeywordField, Schema
            from pydantic import BaseModel

            class ColoredTag(BaseModel):
                name: str   # e.g. of value 'important'
                color: str  # e.g. of value '#ff00dd'

            class Index(Schema):
                id: Annotated[str, IdField(primary_key=True)]
                tags: Annotated[List[ColoredTag], KeywordField()]

        In context of full text search user will
        look up models only based on tags name; in other words, it does
        not make sense to index the color of the tag.
        However, when it comes to displaying the tag to the user, user of course
        wants to see the colored tag (not just names).

        `get_field_value` returns value of the given field which will actually
        be indexed (i.e. user will search by that specific part of the field).
        This method may be overriden by defining method with following name:

            `get_idx_value__<field_name>`
        """
        if hasattr(self, f'get_idx_value__{field_name}'):
            func = getattr(self, f'get_idx_value__{field_name}')
            return func()

        return getattr(self, field_name)



import xapian

from salinic.url import URL
from salinic.utils import get_db_path

RO = 'ro'
RW = 'rw'


class ClientRW:

    def __init__(self, url: URL):
        self._db_path = url.index
        self._db = xapian.WritableDatabase(
            self._db_path,
            xapian.DB_CREATE_OR_OPEN
        )

    def replace_document(self, idterm: str, doc: xapian.Document):
        self.db.replace_document(idterm, doc)
        self.commit()

    def delete_document(self, docid: str):
        self.db.delete_document(docid)
        self.commit()

    def commit(self) -> None:
        self.db.commit()

    @property
    def db(self) -> xapian.Database | xapian.WritableDatabase:
        return self._db


class ClientRO:

    def __init__(self, dsn: str):
        self._db_path = get_db_path(dsn)
        self._db = xapian.Database(self._db_path)

    def replace_document(self, idterm: str, doc: xapian.Document):
        self.db.replace_document(idterm, doc)
        self.commit()

    def delete_document(self, docid: str):
        self.db.delete_document(docid)
        self.commit()

    def commit(self) -> None:
        self.db.commit()

    @property
    def db(self) -> xapian.Database | xapian.WritableDatabase:
        return self._db
